from ._cli import get_config as get_config
from ._cli import main as main
from ._module import Module as Module
from ._module import initialize as initialize
from ._config import get_root_module as get_root_module
from ._config import merge_config as merge_config



import sys

import pytest

from fastaio import Module

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

pytestmark = pytest.mark.anyio


async def test_exception_prepare():
    outputs = []
    error = RuntimeError("prepare0")

    class Module0(Module):
        async def prepare(self):
            outputs.append("prepare0")
            raise error

        async def start(self):
            # should not be called, since prepare failed
            outputs.append("start0")  # pragma: no cover

        async def stop(self):
            # should always be called
            outputs.append("stop0")

    async with Module0("module0") as module0:
        pass

    assert module0.exceptions == [error]
    assert outputs == ["prepare0", "stop0"]


async def test_exception_start():
    outputs = []
    error = RuntimeError("start0")

    class Module0(Module):
        async def prepare(self):
            outputs.append("prepare0")

        async def start(self):
            outputs.append("start0")
            raise error

        async def stop(self):
            # should always be called
            outputs.append("stop0")

    async with Module0("module0") as module0:
        pass

    assert module0.exceptions == [error]


async def test_exception_stop():
    outputs = []
    error = RuntimeError("stop0")

    class Module0(Module):
        async def prepare(self):
            outputs.append("prepare0")

        async def start(self):
            outputs.append("start0")

        async def stop(self):
            outputs.append("stop0")
            raise error

    async with Module0("module0") as module0:
        pass

    assert module0.exceptions == [error]
    assert outputs == ["prepare0", "start0", "stop0"]


async def test_exception_prepare_stop():
    outputs = []
    error_prepare0 = RuntimeError("prepare0")
    error_stop0 = RuntimeError("stop0")

    class Module0(Module):
        async def prepare(self):
            outputs.append("prepare0")
            raise error_prepare0

        async def start(self):
            outputs.append("start0")  # pragma: no cover

        async def stop(self):
            outputs.append("stop0")
            raise error_stop0

    async with Module0("module0") as module0:
        pass

    assert module0.exceptions == [error_prepare0, error_stop0]
    assert outputs == ["prepare0", "stop0"]


async def test_exception_start_stop():
    outputs = []
    error_start0 = RuntimeError("start0")
    error_stop0 = RuntimeError("stop0")

    class Module0(Module):
        async def prepare(self):
            outputs.append("prepare0")

        async def start(self):
            outputs.append("start0")
            raise error_start0

        async def stop(self):
            outputs.append("stop0")
            raise error_stop0

    async with Module0("module0") as module0:
        pass

    assert module0.exceptions == [error_start0, error_stop0]
    assert outputs == ["prepare0", "start0", "stop0"]


async def test_exception_submodule():
    outputs = []
    error_sub_start0 = RuntimeError("sub start0")
    error_sub_stop0 = RuntimeError("sub stop0")

    class Submodule0(Module):
        async def prepare(self):
            outputs.append("sub prepare0")

        async def start(self):
            outputs.append("sub start0")
            raise error_sub_start0

        async def stop(self):
            outputs.append("sub stop0")
            raise error_sub_stop0

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0")

        async def prepare(self):
            outputs.append("prepare0")

        async def start(self):
            outputs.append("start0")

        async def stop(self):
            outputs.append("stop0")

    async with Module0("module0") as module0:
        pass

    assert module0.exceptions == [error_sub_start0, error_sub_stop0]
    assert outputs == [
        "prepare0",
        "sub prepare0",
        "start0",
        "sub start0",
        "stop0",
        "sub stop0",
    ]



import sys

import pytest

from anyio import sleep
from fastaio import Module

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

pytestmark = pytest.mark.anyio


async def test_value():

    value0 = None
    value1 = None

    class Value0:
        pass

    class Value1:
        pass

    class Submodule0(Module):
        async def start(self):
            nonlocal value1
            self.value0 = Value0()
            self.put(self.value0)
            self.value1 = value1 = await self.get(Value1)

        async def stop(self):
            self.drop_value(self.value1)
            await self.value_freed(self.value0)

    class Submodule1(Module):
        async def start(self):
            nonlocal value0
            self.value0 = value0 = await self.get(Value0)
            self.value1 = Value1()
            self.put(self.value1)

        async def stop(self):
            await self.value_freed(self.value1)
            self.drop_value(self.value0)

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0")
            self.add_module(Submodule1, "submodule1")

    async with Module0("module0") as module0:
        pass

    assert type(module0.modules["submodule1"].value0) == Value0
    assert type(module0.modules["submodule0"].value1) == Value1
    assert module0.modules["submodule1"].value0 == value0
    assert module0.modules["submodule0"].value1 == value1


async def test_value_level():

    class Value0:
        pass

    class Value1:
        pass

    class Value2:
        pass

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Module1, "module1")

        async def start(self):
            self.value0 = Value0()
            self.put(self.value0)
            self.value1 = await self.get(Value1, timeout=0.1)
            self.value2 = await self.get(Value2, timeout=0.1)

    class Module1(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Module2, "module2")

        async def start(self):
            self.value1 = Value1()
            self.put(self.value1)
            self.value0 = await self.get(Value0, timeout=0.1)
            self.value2 = await self.get(Value2, timeout=0.1)

    class Module2(Module):
        async def start(self):
            self.value2 = Value2()
            self.put(self.value2)
            self.value0 = await self.get(Value0, timeout=0.1)
            self.value1 = await self.get(Value1, timeout=0.1)

    async with Module0("module0") as module0:
        pass

    module1 = module0.modules["module1"]
    module2 = module1.modules["module2"]
    assert module0.value1 == module1.value1
    assert module0.value2 is None
    assert module1.value0 == module0.value0
    assert module1.value2 == module2.value2
    assert module2.value0 is None
    assert module2.value1 == module1.value1


async def test_get_timeout():

    class Module0(Module):
        async def start(self):
            self.value0 = await self.get(str, timeout=0)

    async with Module0("module0") as module0:
        pass

    assert module0.value0 is None


async def test_value_with_context_manager():
    outputs = []

    class Value0:
        async def __aenter__(self):
            outputs.append("aenter")
            return self

        async def __aexit__(self, exc_type, exc_value, exc_tb):
            outputs.append("aexit")

        def __enter__(self):
            outputs.append("enter")
            return self

        def __exit__(self, exc_type, exc_value, exc_tb):
            outputs.append("exit")

    class Module0(Module):
        async def start(self):
            outputs.append("start")
            value0 = Value0()
            await self.async_context_manager(value0)
            self.context_manager(value0)

        async def stop(self):
            outputs.append("stop")

    async with Module0("module0"):
        pass

    assert outputs == [
        "start",
        "aenter",
        "enter",
        "exit",
        "aexit",
        "stop",
    ]


async def test_put_with_type():

    class Submodule0(Module):
        async def start(self):
            self.put(0, types=int)

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0")

        async def start(self):
            self.value = await self.get(int)
            assert self.value == 0

        async def stop(self):
            self.drop_value(self.value)

    async with Module0("module0"):
        pass


async def test_put_same_value_type():

    class Module0(Module):
        async def start(self):
            self.put(0)
            self.put(0)

    async with Module0("module0") as module0:
        pass

    assert len(module0.exceptions) == 1
    assert type(module0.exceptions[0]) is RuntimeError
    assert str(module0.exceptions[0]) == """Value type "<class 'int'>" already exists"""


async def test_put_exclusive_value():
    outputs = []

    class Submodule0(Module):
        async def start(self):
            self.put(0, exclusive=True)

    class Submodule1(Module):
        async def start(self):
            value = await self.get(int)
            outputs.append("get")
            await sleep(0.1)
            outputs.append("drop")
            self.drop_value(value)

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0")
            self.add_module(Submodule1, "submodule1")
            self.add_module(Submodule1, "submodule2")

    async with Module0("module0"):
        pass

    assert outputs == [
        "get",
        "drop",
        "get",
        "drop",
    ]


async def test_value_not_freed():

    class Submodule0(Module):
        async def start(self):
            self.put(0, types=int)

    class Module0(Module):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.add_module(Submodule0, "submodule0")

        async def start(self):
            self.value = await self.get(int)
            assert self.value == 0

        async def stop(self):
            await sleep(1)

    async with Module0("module0", stop_timeout=0.1) as module0:
        pass

    assert len(module0.exceptions) == 2
    assert str(module0.exceptions[0]) == "Module timed out while stopping: module0.submodule0"
    assert str(module0.exceptions[1]) == "Module timed out while stopping: module0"


async def test_all_values_freed():
    outputs = []

    class Submodule0(Module):
        async def start(self):
            self.put(0, types=int)

        async def stop(self):
            await self.all_values_freed()
            outputs.append("all values freed")

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0")

        async def start(self):
            self.value = await self.get(int)
            assert self.value == 0

        async def stop(self):
            self.drop_value(self.value)
            outputs.append("value dropped")

    async with Module0("module0"):
        pass

    assert outputs == [
        "value dropped",
        "all values freed",
    ]



import pytest

from anyio import create_task_group, sleep
from fastaio import Module

pytestmark = pytest.mark.anyio


async def test_module():
    outputs = []

    class Submodule0(Module):
        async def start(self):
            outputs.append("started0")

        async def stop(self):
            outputs.append("stopped0")

    class Submodule1(Module):
        async def start(self):
            outputs.append("started1")

        async def stop(self):
            outputs.append("stopped1")

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0")
            self.add_module(Submodule1, "submodule1")

    async with Module0("module0") as module0:
        pass

    assert module0.started
    assert outputs in (
        ["started0", "started1", "stopped0", "stopped1"],
        ["started0", "started1", "stopped1", "stopped0"],
        ["started1", "started0", "stopped1", "stopped0"],
        ["started1", "started0", "stopped0", "stopped1"],
    )


async def test_add_module_str():
    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module("fastaio:Module", "submodule0")

    async with Module0("module0") as module0:
        pass

    assert type(module0.modules["submodule0"]) is Module


async def test_add_same_module_name():
    class Submodule0(Module):
        pass

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0")
            self.add_module(Submodule0, "submodule0")

    with pytest.raises(RuntimeError) as excinfo:
        Module0("module0")

    assert str(excinfo.value) == "Module name already exists: submodule0"


async def test_module_not_initialized():
    class Module0(Module):
        def __init__(self, name):
            pass

    module0 = Module0("module0")

    with pytest.raises(RuntimeError) as excinfo:
        async with module0:
            pass  # pragma: no cover

    assert str(excinfo.value) == "You must call super().__init__() in the __init__ method of your module"



import pytest

import httpx
from anyio import connect_tcp, create_task_group
from fastapi import FastAPI
from fastapi.testclient import TestClient

from fastaio import Module
from fastaio.web.fastapi import FastAPIModule

pytestmark = pytest.mark.anyio


async def test_web(unused_tcp_port):
    class Submodule0(Module):
        async def prepare(self):
            self.app = await self.get(FastAPI)

        async def start(self):
            @self.app.get("/")
            def read_root():
                return {"Hello": "World"}

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(FastAPIModule, "fastapi_module", port=unused_tcp_port)
            self.add_module(Submodule0, "submodule0")

    async with Module0("module0") as module0:
            app = module0.modules["submodule0"].app
            async with httpx.AsyncClient() as client:
                response = await client.get(f"http://127.0.0.1:{unused_tcp_port}")

    assert response.json() == {"Hello": "World"}



import sys

import pytest

from anyio import TASK_STATUS_IGNORED, create_task_group, sleep
from anyio.abc import TaskStatus
from fastaio import Module

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

pytestmark = pytest.mark.anyio


async def test_task():
    outputs = []

    class Value0:
        pass

    async def task(message: str, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        outputs.append(message)
        task_status.started()
        await sleep(float("inf"))

    class Submodule0(Module):
        async def prepare(self):
            async with create_task_group() as self.tg1:
                await self.tg1.start(task, "prepare0", name="prepare0")
                self.done()
                outputs.append("prepared0")

        async def start(self):
            async with create_task_group() as self.tg2:
                await self.tg2.start(task, "start0", name="start0")
                self.done()
                outputs.append("started0")
                self.value0 = Value0()
                self.put(self.value0)

        async def stop(self):
            await self.value_freed(self.value0)
            self.tg1.cancel_scope.cancel()
            self.tg2.cancel_scope.cancel()
            outputs.append("stopped0")

    class Submodule1(Module):
        async def start(self):
            self.value0 = await self.get(Value0)
            async with create_task_group() as self.tg:
                await self.tg.start(task, "start1", name="start1")
                self.done()
                outputs.append("started1")

        async def stop(self):
            self.tg.cancel_scope.cancel()
            outputs.append("stopped1")
            self.drop_value(self.value0)

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0")
            self.add_module(Submodule1, "submodule1")

    module0 = Module0("module0")

    async with module0:
        pass

    assert outputs == [
        "prepare0",
        "prepared0",
        "start0",
        "started0",
        "start1",
        "started1",
        "stopped1",
        "stopped0",
    ]


async def test_failing_task():
    outputs = []

    async def failing_task():
        await sleep(0.05)
        raise RuntimeError("start0")

    class Submodule0(Module):
        async def start(self):
            async with create_task_group() as self.tg:
                self.tg.start_soon(failing_task)
                outputs.append("started0")

    class Submodule1(Module):
        pass

    class Module0(Module):
        def __init__(self, name):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0")
            self.add_module(Submodule1, "submodule1")

    async with Module0("module0") as module0:
        await sleep(0.1)
    
    assert len(module0.exceptions) == 1
    assert type(module0.exceptions[0]) is ExceptionGroup
    assert str(module0.exceptions[0].exceptions[0]) == "start0"

    assert outputs == [
        "started0",
    ]



import click

from ._importer import import_from_string


@click.command()
@click.option("--set", "set_", multiple=True, help="The assignment to the module parameter.")
@click.argument("module")
def main(module, set_):
    global CONFIG
    module_type = import_from_string(module)
    config = {
        "root_module": {
            "type": module_type,
        }
    }
    for _set in set_:
        if "=" not in _set:
            raise click.ClickException(f"No '=' while setting a module parameter: {_set}")

        key, value = _set.split("=", 1)
        path = key.split(".")
        modules = config["root_module"]
        for module_name in path[:-1]:
            modules = modules.setdefault("modules", {})
            modules = modules.setdefault(module_name, {})
        _config = modules.setdefault("config", {})
        _config[path[-1]] = value
    CONFIG = config


def get_config():
    return CONFIG



import pytest

from fastaio import Module, get_root_module, initialize, merge_config


def test_config_override():
    class Submodule0(Module):
        def __init__(self, name, param0="param0", param1="param1"):
            super().__init__(name)
            self.param0 = param0
            self.param1 = param1
    
    class Submodule1(Module):
        def __init__(self, name, param0="param0"):
            super().__init__(name)
            self.param0 = param0

    class Submodule2(Module):
        def __init__(self, name, param2="param2"):
            super().__init__(name)
            self.add_module(Submodule3, "submodule3", param3="param3*")
            self.param2 = param2

    class Submodule3(Module):
        def __init__(self, name, param3="param3"):
            super().__init__(name)
            self.add_module(Submodule4, "submodule4", param4="param4*")
            self.param3 = param3

    class Submodule4(Module):
        def __init__(self, name, param4="param4"):
            super().__init__(name)
            self.param4 = param4

    class Module0(Module):
        def __init__(self, name, param0="param0"):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0", param0="param0*")
            self.add_module(Submodule1, "submodule1")
            self.add_module(Submodule2, "submodule2", param2="param2*")
            self.param0 = param0

    module0 = Module0("module0", param0="bar")
    initialize(module0)
    initialize(module0)
    assert module0.param0 == "bar"
    assert module0.modules["submodule0"].param0 == "param0*"
    assert module0.modules["submodule0"].param1 == "param1"
    assert module0.modules["submodule1"].param0 == "param0"
    assert module0.modules["submodule2"].param2 == "param2*"
    assert module0.modules["submodule2"].modules["submodule3"].param3 == "param3*"
    assert module0.modules["submodule2"].modules["submodule3"].modules["submodule4"].param4 == "param4*"


def test_config_from_dict():
    class Submodule0(Module):
        def __init__(self, name, param0="param0", param1="param1"):
            super().__init__(name)
            self.param0 = param0
            self.param1 = param1
    
    class Submodule1(Module):
        def __init__(self, name, param0="param0"):
            super().__init__(name)
            self.param0 = param0

    class Submodule2(Module):
        def __init__(self, name, param2="param2"):
            super().__init__(name)
            self.add_module(Submodule3, "submodule3")
            self.param2 = param2

    class Submodule3(Module):
        def __init__(self, name, param3="param3"):
            super().__init__(name)
            self.add_module(Submodule4, "submodule4", param4="foo")
            self.param3 = param3

    class Submodule4(Module):
        def __init__(self, name, param4="param4"):
            super().__init__(name)
            self.param4 = param4

    class Module0(Module):
        def __init__(self, name, param0="param0"):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0", param0="foo")
            self.add_module(Submodule1, "submodule1")
            self.add_module(Submodule2, "submodule2")
            self.param0 = param0

    config = {
        "module0": {
            "type": Module0,
            "config": {
                "param0": "bar",
            },
            "modules": {
                "submodule0": {
                    "config": {
                        "param0": "foo2",
                    },
                },
                "submodule1": {
                    "config": {
                        "param0": "baz",
                    },
                },
                "submodule2": {
                    "config": {
                        "param2": "param2*",
                    },
                    "modules": {
                        "submodule3": {
                            "config": {
                                "param3": "param3*",
                            },
                            "modules": {
                                "submodule4": {
                                    "config": {
                                        "param4": "param4*",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
    }

    module0 = get_root_module(config)
    initialize(module0)
    assert module0.param0 == "bar"
    assert module0.modules["submodule0"].param0 == "foo2"
    assert module0.modules["submodule0"].param1 == "param1"
    assert module0.modules["submodule1"].param0 == "baz"
    assert module0.modules["submodule2"].param2 == "param2*"
    assert module0.modules["submodule2"].modules["submodule3"].param3 == "param3*"
    assert module0.modules["submodule2"].modules["submodule3"].modules["submodule4"].param4 == "param4*"


def test_config_from_dict_with_type_as_str():
    config = {
        "module0": {
            "type": "fastaio:Module",
            "modules": {
                "module0": {
                    "type": "fastaio:Module",
                    "modules": {
                        "module00": {
                            "type": "fastaio:Module",
                        },
                    },
                },
            },
        },
    }

    module0 = get_root_module(config)
    initialize(module0)
    assert module0.modules["module0"].modules["module00"]


def test_wrong_config_from_dict_1():
    class Module0(Module):
        pass

    config = {
        "module0": {
            "type": Module0,
            "modules": {
                "submodule0": {
                    "config": {
                        "param0": "foo",
                    },
                },
            },
        },
    }

    module0 = get_root_module(config)

    with pytest.raises(RuntimeError) as excinfo:
        initialize(module0)

    assert str(excinfo.value) == "Module not found: submodule0"


def test_wrong_config_from_dict_2():
    class Submodule0(Module):
        def __init__(self, name, param0="param0", param1="param1"):
            super().__init__(name)
            self.param0 = param0
            self.param1 = param1

    class Module0(Module):
        def __init__(self, name, param0="param0"):
            super().__init__(name)
            self.add_module(Submodule0, "submodule0", param0="foo")

    config = {
        "module0": {
            "type": Module0,
            "modules": {
                "submodule0": {
                    "config": {
                        "param0": "foo",
                    },
                    "modules": {
                        "submodule1": {
                            "config": {
                                "param1": "bar",
                            },
                        },
                    },
                },
            },
        },
    }

    module0 = get_root_module(config)

    with pytest.raises(RuntimeError) as excinfo:
        initialize(module0)

    assert str(excinfo.value) == "Module not found: submodule1"


def test_config_from_dict_add_submodules():
    class Submodule0(Module):
        def __init__(self, name, param0="param0", param1="param1"):
            super().__init__(name)
            self.param0 = param0
            self.param1 = param1

    class Submodule1(Module):
        def __init__(self, name, param0="param0", param1="param1"):
            super().__init__(name)
            self.param0 = param0
            self.param1 = param1

    class Submodule10(Module):
        def __init__(self, name, param0="param0", param1="param1"):
            super().__init__(name)
            self.param0 = param0
            self.param1 = param1

    class Module0(Module):
        pass

    config = {
        "module0": {
            "type": Module0,
            "modules": {
                "submodule0": {
                    "type": Submodule0,
                    "config": {
                        "param0": "foo",
                    },
                },
                "submodule1": {
                    "type": Submodule1,
                    "config": {
                        "param1": "bar",
                    },
                    "modules": {
                        "submodule10": {
                            "type": Submodule10,
                            "config": {
                                "param0": "baz",
                             },
                        },
                    },
                },
            },
        },
    }

    module0 = get_root_module(config)
    initialize(module0)
    assert list(module0.modules.keys()) == ["submodule0", "submodule1"]
    assert list(module0.modules["submodule0"].modules.keys()) == []
    assert list(module0.modules["submodule1"].modules.keys()) == ["submodule10"]
    assert list(module0.modules["submodule1"].modules["submodule10"].modules.keys()) == []
    assert module0.modules["submodule0"].param0 == "foo"
    assert module0.modules["submodule0"].param1 == "param1"
    assert module0.modules["submodule1"].param0 == "param0"
    assert module0.modules["submodule1"].param1 == "bar"
    assert module0.modules["submodule1"].modules["submodule10"].param0 == "baz"
    assert module0.modules["submodule1"].modules["submodule10"].param1 == "param1"


def test_merge_config():
    d0 = {
        "module0": {
            "type": "Module0",
            "config": {
                "param0": 0,
                "param1": 1,
            },
            "modules": {
                "module1": {
                    "type": "Module1",
                    "modules": {
                        "module2": {
                            "type": "Module2",
                            "config": {
                                "param2": 2,
                                "param3": 3,
                            }
                        }
                    }
                }
            }
        }
    }

    d1 = {
        "module0": {
            "config": {
                "param1": 11,
            },
            "modules": {
                "module1": {
                    "modules": {
                        "module2": {
                            "config": {
                                "param2": 22,
                            },
                        },
                        "module3": {
                            "type": "Module3",
                            "config": {
                                "param4": 4,
                            }
                        }
                    }
                }
            }
        }
    }

    d = merge_config(d0, d1)
    assert d == {
        "module0": {
            "type": "Module0",
            "config": {
                "param0": 0,
                "param1": 11,
            },
            "modules": {
                "module1": {
                    "type": "Module1",
                    "modules": {
                        "module2": {
                            "type": "Module2",
                            "config": {
                                "param2": 22,
                                "param3": 3,
                            }
                        },
                        "module3": {
                            "type": "Module3",
                            "config": {
                                "param4": 4,
                            }
                        }
                    }
                }
            }
        }
    }



import importlib
import sys
from typing import Any

if sys.version_info < (3, 10):  # pragma: nocover
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


class ImportFromStringError(Exception):
    pass


def import_from_string(import_str: Any) -> Any:
    if not isinstance(import_str, str):
        return import_str

    if ":" not in import_str:
        # this is an entry-point in the "fastaio.modules" group
        for ep in entry_points(group="fastaio.modules"):
            if ep.name == import_str:
                return ep.load()
        raise RuntimeError(f'Module could not be found in entry-point group "fastaio.modules": {import_str}')

    module_str, _, attrs_str = import_str.partition(":")
    try:
        module = importlib.import_module(module_str)
    except ModuleNotFoundError as exc:
        if exc.name != module_str:  # pragma: nocover
            raise exc from None
        message = 'Could not import module "{module_str}".'
        raise ImportFromStringError(message.format(module_str=module_str))

    instance = module
    try:
        for attr_str in attrs_str.split("."):
            instance = getattr(instance, attr_str)
    except AttributeError:
        message = 'Attribute "{attrs_str}" not found in module "{module_str}".'
        raise ImportFromStringError(message.format(attrs_str=attrs_str, module_str=module_str))

    return instance



from __future__ import annotations

import logging
import sys

from contextlib import AsyncExitStack
from inspect import isawaitable, signature
from typing import TYPE_CHECKING, TypeVar, Any, Callable, Iterable

import anyio
import structlog
from anyio import Event, create_task_group, move_on_after
from anyioutils import create_task, wait, FIRST_COMPLETED

from ._container import Container
from ._importer import import_from_string

if TYPE_CHECKING:
    from ._container import Value  # pragma: no cover

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

log = structlog.get_logger()
structlog.stdlib.recreate_defaults(log_level=logging.INFO)

T_Value = TypeVar("T_Value")


class Module:
    _exit: Event
    _exceptions: list[Exception]

    def __init__(
        self,
        name: str,
        prepare_timeout=1,
        start_timeout=1,
        stop_timeout=1,
    ):
        self._initialized = False
        self._prepare_timeout = prepare_timeout
        self._start_timeout = start_timeout
        self._stop_timeout = stop_timeout
        self._parent: Module | None = None
        self._container = Container()
        self._prepared = Event()
        self._started = Event()
        self._stopped = Event()
        self._name = name
        self._path: list[str] = []
        self._uninitialized_modules: dict[str, Any] = {}
        self._modules: dict[str, Module] = {}
        self._added_values: dict[Any, Value] = {}
        self._acquired_values: dict[Any, Value] = {}
        self._context_manager_exits: list[Callable] = []

    @property
    def parent(self) -> Module | None:
        return self._parent

    @parent.setter
    def parent(self, value: Module) -> None:
        self._parent = value
        self._exit = value._exit

    @property
    def path(self) -> str:
        return ".".join(self._path + [self._name])

    @property
    def started(self) -> Event:
        return self._started

    @property
    def exceptions(self) -> list[Exception]:
        return self._exceptions

    def _check_init(self):
        try:
            self._initialized
        except AttributeError:
            raise RuntimeError("You must call super().__init__() in the __init__ method of your module")

    @property
    def modules(self) -> dict[str, Module]:
        return self._modules

    def add_module(
        self,
        module_type: type["Module"] | str,
        name: str,
        **config,
    ) -> None:
        self._check_init()
        if name in self._uninitialized_modules:
            raise RuntimeError(f"Module name already exists: {name}")
        module_type = import_from_string(module_type)
        self._uninitialized_modules[name] = {
            "type": module_type,
            "config": config,
            "modules": {},
        }
        log.debug("Module added", path=self.path, name=name, module_type=module_type)

    async def value_freed(self, value: Any) -> None:
        value_id = id(value)
        await self._added_values[value_id].wait_no_borrower()

    async def all_values_freed(self) -> None:
        for value in self._added_values.values():
            await value.wait_no_borrower()

    def drop_all_values(self) -> None:
        for value in self._acquired_values.values():
            value.drop(self)

    def drop_value(self, value: Any) -> None:
        value_id = id(value)
        self._acquired_values[value_id].drop(self)

    def put(self, value: T_Value, types: Iterable | Any | None = None, exclusive: bool = False) -> None:
        value_id = id(value)
        value_types = self._container.get_value_types(value, types)
        self._added_values[value_id] = self._container.put(value, self, value_types, exclusive)
        if self.parent is not None:
            self._added_values[value_id] = self.parent._container.put(value, self, value_types, exclusive)
        log.debug("Module added value", path=self.path, value_types=value_types)

    async def get(self, value_type: T_Value, timeout: float = float("inf")) -> T_Value | None:
        log.debug("Module getting value", path=self.path, value_type=value_type)
        tasks = [create_task(self._container.get(value_type, self), self._task_group)]
        if self.parent is not None:
            tasks.append(create_task(self.parent._container.get(value_type, self), self._task_group))
        with move_on_after(timeout) as scope:
            done, pending = await wait(tasks, self._task_group, return_when=FIRST_COMPLETED)
            for task in pending:
                task.cancel(raise_exception=False)
            for task in done:
                break
            value = await task.wait()
        if scope.cancelled_caught:
            log.debug("Module did not get value in time", path=self.path, value_type=value_type)
            return None
        value_id = id(value._value)
        self._acquired_values[value_id] = value
        log.debug("Module got value", path=self.path, value_type=value_type)
        return value._value

    async def __aenter__(self) -> Module:
        self._check_init()
        log.debug("Running root module", name=self.path)
        initialize(self)
        async with AsyncExitStack() as exit_stack:
            self._task_group = await exit_stack.enter_async_context(create_task_group())
            self._exceptions = []
            self._phase = "preparing"
            with move_on_after(self._prepare_timeout) as scope:
                self._task_group.start_soon(self._prepare, name=f"{self.path} _prepare")
                await self._all_prepared()
            if scope.cancelled_caught:
                self._get_all_prepare_timeout()
            if self._exceptions:
                self._exit.set()
            else:
                self._phase = "starting"
                with move_on_after(self._start_timeout) as scope:
                    self._task_group.start_soon(self._start, name=f"{self.path} start")
                    await self._all_started()
                if scope.cancelled_caught:
                    self._get_all_start_timeout()
                if self._exceptions:
                    self._exit.set()
                if not self._exit.is_set():
                    log.debug("Application running")
            self._exit_stack = exit_stack.pop_all()
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        self._phase = "stopping"
        with move_on_after(self._stop_timeout) as scope:
            self._task_group.start_soon(self._stop, name=f"{self.path} stop")
            await self._all_stopped()
        self._exit.set()
        if scope.cancelled_caught:
            self._get_all_stop_timeout()
        self._task_group.cancel_scope.cancel()
        try:
            await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)
        except anyio.get_cancelled_exc_class():  # pragma: nocover
            pass
        exceptions = []
        for exc in self._exceptions:
            while isinstance(exc, ExceptionGroup):
                exc = exc.exceptions[0]
            exceptions.append(exc)
        if exceptions:
            log.critical("Application failed")
            for exception in exceptions:
                log.critical("Exception", exc_info=exception)
        log.debug("Application stopped")

    def context_manager(self, value):
        self._context_manager_exits.append(value.__exit__)
        return value.__enter__()

    async def async_context_manager(self, value):
        self._context_manager_exits.append(value.__aexit__)
        return await value.__aenter__()

    def _get_all_prepare_timeout(self):
        for module in self._modules.values():
            module._get_all_prepare_timeout()
        if not self._prepared.is_set():
            self._exceptions.append(TimeoutError(f"Module timed out while preparing: {self.path}"))

    def _get_all_start_timeout(self):
        for module in self._modules.values():
            module._get_all_start_timeout()
        if not self._started.is_set():
            self._exceptions.append(TimeoutError(f"Module timed out while starting: {self.path}"))

    def _get_all_stop_timeout(self):
        for module in self._modules.values():
            module._get_all_stop_timeout()
        if not self._stopped.is_set():
            self._exceptions.append(TimeoutError(f"Module timed out while stopping: {self.path}"))

    async def _all_prepared(self):
        for module in self._modules.values():
            await module._all_prepared()
        await self._prepared.wait()

    async def _all_started(self):
        for module in self._modules.values():
            await module._all_started()
        await self._started.wait()

    async def _all_stopped(self):
        for module in self._modules.values():
            await module._all_stopped()
        await self._stopped.wait()

    async def _prepare(self) -> None:
        log.debug("Preparing module", path=self.path)
        try:
            async with create_task_group() as tg:
                for module in self._modules.values():
                    module._task_group = tg
                    module._phase = self._phase
                    module._exceptions = self._exceptions
                    tg.start_soon(module._prepare, name=f"{module.path} _prepare")
                tg.start_soon(self._prepare_and_done, name=f"{self.path} _prepare_and_done")
        except ExceptionGroup as exc:
            self._exceptions.append(*exc.exceptions)
            self._prepared.set()
            log.debug("Module failed while preparing", path=self.path, exc_info=exc)

    async def _prepare_and_done(self) -> None:
        await self.prepare()
        self.done()

    async def prepare(self) -> None:
        pass

    def done(self) -> None:
        if self._phase == "preparing":
            self._prepared.set()
            log.debug("Module prepared", path=self.path)
        elif self._phase == "starting":
            self._started.set()
            log.debug("Module started", path=self.path)
        else:
            self._task_group.start_soon(self._finish)

    async def _finish(self):
        tasks = (
            create_task(self._drop_and_wait_values(), self._task_group),
            create_task(self._exit.wait(), self._task_group),
        )
        done, pending = await wait(tasks, self._task_group, return_when=FIRST_COMPLETED)
        for task in pending:
            task.cancel(raise_exception=False)

    async def _drop_and_wait_values(self):
        self.drop_all_values()
        await self.all_values_freed()
        self._stopped.set()
        log.debug("Module stopped", path=self.path)

    async def _start(self) -> None:
        log.debug("Starting module", path=self.path)
        try:
            async with create_task_group() as tg:
                for module in self._modules.values():
                    module._task_group = tg
                    module._phase = self._phase
                    tg.start_soon(module._start, name=f"{module.path} _start")
                tg.start_soon(self._start_and_done, name=f"{self.path} _start_and_done")
        except ExceptionGroup as exc:
            self._exceptions.append(*exc.exceptions)
            self._started.set()

    async def _start_and_done(self) -> None:
        await self.start()
        self.done()

    async def start(self) -> None:
        pass

    async def _stop(self) -> None:
        log.debug("Stopping module", path=self.path)
        try:
            async with create_task_group() as tg:
                for module in self._modules.values():
                    module._task_group = tg
                    module._phase = self._phase
                    tg.start_soon(module._stop, name=f"{module.path} _stop")
                for context_manager_exit in self._context_manager_exits[::-1]:
                    res = context_manager_exit(None, None, None)
                    if isawaitable(res):
                        await res   
                tg.start_soon(self._stop_and_done, name=f"{self.path} _stop_and_done")
        except ExceptionGroup as exc:
            self._exceptions.append(*exc.exceptions)
            self._stopped.set()

    async def _stop_and_done(self) -> None:
        await self.stop()
        self.done()

    async def stop(self) -> None:
        pass

    async def _main(self): # pragma: no cover
        async with self:
            await self._exit.wait()

    def run(self):  # pragma: no cover
        try:
            anyio.run(self._main)
        except KeyboardInterrupt:
            pass


def initialize(root_module: Module) -> None:
    if root_module._initialized:
        return

    root_module._exit = Event()
    _initialize(root_module._uninitialized_modules, root_module, root_module._uninitialized_modules)
    root_module._uninitialized_modules = {}
    root_module._initialized = True


def _initialize(submodules: dict[str, Any], parent_module: Module, root_module_modules: dict[str, Any]) -> None:
    for name, info in root_module_modules.items():
        if name in submodules:
            if info.get("type") is not None:
                submodules[name]["type"] = info["type"]
        else:
            submodules[name] = info
    for name, info in submodules.items():
        config = info.get("config", {})
        config.update(root_module_modules.get(name, {}).get("config", {}))
        if "type" not in info:
            raise RuntimeError(f"Module not found: {name}")
        module_type = import_from_string(info["type"])
        submodule_instance: Module = module_type(name, **config)
        submodule_instance._path = parent_module._path + [parent_module._name]
        submodule_instance.parent = parent_module
        parent_module._modules[name] = submodule_instance
        _initialize(submodule_instance._uninitialized_modules, submodule_instance, root_module_modules.get(name, {}).get("modules", {}))
        submodule_instance._uninitialized_modules = {}



import sys

import pytest

from anyio import sleep
from fastaio import Module

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

pytestmark = pytest.mark.anyio


async def test_prepare():

    class Module0(Module):
        async def prepare(self):
            # will never prepare
            await sleep(1)

    async with Module0("module0", prepare_timeout=0.1) as module0:
        pass

    assert len(module0.exceptions) == 1
    assert str(module0.exceptions[0]) == "Module timed out while preparing: module0"


async def test_nested_prepare():

    class Submodule0(Module):
        async def prepare(self):
            # will never prepare
            await sleep(1)

    class Module0(Module):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.submodule0 = self.add_module(Submodule0, "submodule0")

    async with Module0("module0", prepare_timeout=0.1) as module0:
        pass

    assert len(module0.exceptions) == 1
    assert str(module0.exceptions[0]) == "Module timed out while preparing: module0.submodule0"


async def test_start():

    class Module0(Module):
        async def start(self):
            # will never start
            await sleep(1)

    async with Module0("module0", start_timeout=0.1) as module0:
        pass

    assert len(module0.exceptions) == 1
    assert str(module0.exceptions[0]) == "Module timed out while starting: module0"


async def test_nested_start():

    class Submodule0(Module):
        async def start(self):
            # will never start
            await sleep(1)

    class Module0(Module):
        def __init__(self, name, start_timeout):
            super().__init__(name, start_timeout=start_timeout)
            self.submodule0 = self.add_module(Submodule0, "submodule0")

    async with Module0("module0", start_timeout=0.1) as module0:
        pass

    assert len(module0.exceptions) == 1
    assert str(module0.exceptions[0]) == "Module timed out while starting: module0.submodule0"


async def test_stop():

    class Module0(Module):
        async def stop(self):
            # will never stop
            await sleep(1)

    module0 = Module0("module0", stop_timeout=0.1)

    async with module0:
        pass

    assert len(module0.exceptions) == 1
    assert str(module0.exceptions[0]) == "Module timed out while stopping: module0"

async def test_nested_stop():

    class Submodule0(Module):
        async def stop(self):
            # will never stop
            await sleep(1)

    class Module0(Module):
        def __init__(self, name, stop_timeout):
            super().__init__(name, stop_timeout=stop_timeout)
            self.submodule0 = self.add_module(Submodule0, "submodule0")

    async with Module0("module0", stop_timeout=0.1) as module0:
        pass

    assert len(module0.exceptions) == 1
    assert str(module0.exceptions[0]) == "Module timed out while stopping: module0.submodule0"



from __future__ import annotations

from functools import partial

from anyio import Event, connect_tcp, create_task_group
from anyioutils import create_task
from anycorn import Config, serve
from fastapi import FastAPI

from fastaio import Module


class FastAPIModule(Module):
    def __init__(
        self,
        name: str,
        *,
        app: FastAPI | None = None,
        host: str = "127.0.0.1",
        port: int = 8000,
        debug: bool | None = None,
    ) -> None:
        super().__init__(name)
        debug = debug if debug is not None else __debug__
        self.app = app if app is not None else FastAPI(debug=debug)
        self.host = host
        self.port = port
        self.shutdown_event = Event()

    async def prepare(self) -> None:
        self.put(self.app)

    async def start(self) -> None:
        config = Config()
        config.bind = [f"{self.host}:{self.port}"]
        config.loglevel = "WARN"
        async with create_task_group() as tg:
            self.server_task = create_task(serve(self.app, config, shutdown_trigger=self.shutdown_event.wait, mode="asgi"), tg)  # type: ignore[arg-type]
            while True:
                try:
                    await connect_tcp(self.host, self.port)
                except OSError:
                    pass
                else:
                    break
            self.done()

    async def stop(self) -> None:
        self.shutdown_event.set()
        await self.server_task.wait()



from click.testing import CliRunner
from fastaio import Module, get_config, main


def test_wrong_cli_1():
    runner = CliRunner()
    result = runner.invoke(main, [
        "fastaio:Module",
        "--set", "param",
    ])
    assert result.exit_code == 1

def test_wrong_cli_2():
    runner = CliRunner()
    result = runner.invoke(main, [
        "fastaio.Module",
    ])
    assert result.exit_code == 1

def test_wrong_cli_3():
    runner = CliRunner()
    result = runner.invoke(main, [
        "fastaio:WrongModule",
    ])
    assert result.exit_code == 1

def test_wrong_cli_4():
    runner = CliRunner()
    result = runner.invoke(main, [
        "wrong_module:Module",
    ])
    assert result.exit_code == 1

def test_cli():
    runner = CliRunner()
    runner.invoke(main, [
        "fastaio_module",
        "--set", "param=-1",
        "--set", "module0.param0=foo",
        "--set", "module1.param1=bar",
        "--set", "module2.param2=baz",
        "--set", "module2.module3.param3=123",
        ]
    )
    assert get_config() == {
        "root_module": {
            "type": Module,
            "config": {
                "param": "-1"
            },
            "modules": {
                "module0": {
                    "config": {
                        "param0": "foo"
                    }
                },
                "module1": {
                    "config": {
                        "param1": "bar"
                    }
                },
                "module2": {
                    "config": {
                        "param2": "baz"
                    },
                    "modules": {
                        "module3": {
                            "config": {
                                "param3": "123"
                            }
                        }
                    }
                }
            }
        }
    }



from copy import deepcopy
from typing import Any

from ._module import Module
from ._importer import import_from_string


def get_root_module(config: dict[str, Any]) -> Module:
    for module_name, module_info in config.items():
        module_config = module_info.get("config", {})
        module_type = import_from_string(module_info["type"])
        root_module = module_type(module_name, **module_config)
        submodules = module_info.get("modules", {})
        for submodule_name, submodule_info in submodules.items():
            submodule_config = root_module._uninitialized_modules.setdefault(submodule_name, {}).setdefault("config", {})
            submodule_config.update(submodule_info.get("config", {}))
            submodule_type = submodule_info.get("type")
            if submodule_type is not None:
                root_module._uninitialized_modules[submodule_name]["type"] = submodule_type
            root_module._uninitialized_modules[submodule_name]["modules"] = submodule_info.get("modules", {})
        break
    return root_module


def merge_config(config: dict[str, Any], override: dict[str, Any], root: bool = True) -> dict[str, Any]:
    if root:
        config = deepcopy(config)
    for key, val in override.items():
        if key in config:
            if isinstance(val, dict):
                config[key] = merge_config(config[key], override[key], root=False)
            else:
                config[key] = override[key]
        else:
            config[key] = val
    return config



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable

from anyio import Event


if TYPE_CHECKING:
    from ._module import Module  # pragma: no cover


class Value:
    def __init__(self, value: Any, owner: "Module", exclusive: bool):
        self._value = value
        self._owner = owner
        self._exclusive = exclusive
        self._borrowers: set[Module] = set()
        self._dropped = Event()

    def drop(self, borrower: "Module"):
        if borrower in self._borrowers:
            self._borrowers.remove(borrower)
            self._dropped.set()
            self._dropped = Event()

    async def wait_no_borrower(self):
        while True:
            if not self._borrowers:
                return
            await self._dropped.wait()


class Container:
    def __init__(self):
        self._context = {}
        self._value_added = Event()

    def get_value_types(self, value: Any, types: Iterable | Any | None = None) -> Iterable:
        types = types if types is not None else [type(value)]
        try:
            for value_type in types:
                break
        except TypeError as e:
            types = [types]
        return types

    def put(self, value: Any, owner: "Module", types: Iterable, exclusive: bool = False) -> Value:
        _value = Value(value, owner, exclusive)
        for value_type in types:
            value_type_id = id(value_type)
            if value_type_id in self._context:
                raise RuntimeError(f'Value type "{value_type}" already exists')
            self._context[value_type_id] = _value
            self._value_added.set()
            self._value_added = Event()
        return _value

    async def get(self, value_type: Any, borrower: "Module") -> Value:
        value_type_id = id(value_type)
        while True:
            if value_type_id in self._context:
                value = self._context[value_type_id]
                if value._exclusive:
                    while True:
                        if not value._borrowers:
                            value._borrowers.add(borrower)
                            return value
                        await value.wait_no_borrower()
                value._borrowers.add(borrower)
                return value
            await self._value_added.wait()
